baseURLs = [
  ['GLib', 'https://docs.gtk.org/glib/'],
  ['GObject', 'https://docs.gtk.org/gobject/'],
  ['Gio', 'https://docs.gtk.org/gio/'],
  ['Gst', 'https://gstreamer.freedesktop.org/documentation/gstreamer/gi-index.html?'],
  ['GstBase', 'https://gstreamer.freedesktop.org/documentation/base/?'],
  ['GstAudio', 'https://gstreamer.freedesktop.org/documentation/audio/?'],
  ['GstTag', 'https://gstreamer.freedesktop.org/documentation/tag/?'],
  ['GstPbutils', 'https://gstreamer.freedesktop.org/documentation/pbutils/?'],
]
