baseURLs = [
  ['GLib', 'https://docs.gtk.org/glib/'],
  ['GObject', 'https://docs.gtk.org/gobject/'],
  ['Gio', 'https://docs.gtk.org/gio/'],
  ['Gtk', 'https://docs.gtk.org/gtk4/'],
  ['Gst', 'https://gstreamer.freedesktop.org/documentation/gstreamer/gi-index.html?'],
  ['Clapper', 'https://rafostar.github.io/clapper/doc/clapper/'],
]
