/* Clapper GTK Integration Library
 * Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#define __CLAPPER_GTK_INSIDE__

#include <clapper-gtk/clapper-gtk-enums.h>
#include <clapper-gtk/clapper-gtk-version.h>

#include <clapper-gtk/clapper-gtk-billboard.h>
#include <clapper-gtk/clapper-gtk-container.h>
#include <clapper-gtk/clapper-gtk-extra-menu-button.h>
#include <clapper-gtk/clapper-gtk-lead-container.h>
#include <clapper-gtk/clapper-gtk-next-item-button.h>
#include <clapper-gtk/clapper-gtk-previous-item-button.h>
#include <clapper-gtk/clapper-gtk-seek-bar.h>
#include <clapper-gtk/clapper-gtk-simple-controls.h>
#include <clapper-gtk/clapper-gtk-title-header.h>
#include <clapper-gtk/clapper-gtk-title-label.h>
#include <clapper-gtk/clapper-gtk-toggle-fullscreen-button.h>
#include <clapper-gtk/clapper-gtk-toggle-play-button.h>
#include <clapper-gtk/clapper-gtk-utils.h>
#include <clapper-gtk/clapper-gtk-video.h>

#undef __CLAPPER_GTK_INSIDE__
